from dataclasses import asdict, dataclass, field
from time import time


@dataclass
class Comment:
    user: str  #
    body: str
    foreign_key: str
    created: int = field(init=False)
    key: str = field(init=False)

    def __post_init__(self):
        self.created = int(time())
        self.key = f"searches/{self.search_key}/comments/{self.created}"



from dataclasses import asdict, dataclass, field
from typing import List

import pandas as pd

from gandai.datastore import Cloudstore
from gandai.models import Event
from gandai.services import Query

ds = Cloudstore()

# @dataclass
class Search:
    def __init__(self, key) -> None:
        self.key = key

    # client_key: str
    # start_date: str # iso_date
    # strategy: str  # "$4m health food manufacturing for pets"

    # def __post_init__(self):
    # self.key = f"searches/{self.search_key}/events/{self.created}"


# def load_targets(key: str) -> Search:


@dataclass
class SearchData:
    search_key: str


def load_search(search_key: str, actor_key: str) -> SearchData:
    Event.post_event(
        actor_key=actor_key, search_key=search_key, domain="", type="search"
    )
    companies = Query.companies_query(search_key)
    events = Query.events_query(search_key)
    df = companies.merge(
        events[["domain", "type"]], left_on="domain", right_on="domain", how="left"
    )

    def _json_safe(df):
        return df.fillna("").to_dict(orient="records")

    inbox = _json_safe(df[df["type"].isna()])
    reviewed = _json_safe(df[df["type"] == "advance"])
    qualified = _json_safe(df[df["type"] == "qualify"])
    rejected = _json_safe(df[df["type"] == "reject"])

    resp = {
        "search_key": search_key,
        "actor_key": actor_key,
        "meta": {
            "inbox": len(inbox),
            "review": len(reviewed),
            "qualified": len(qualified),
            "rejected": len(rejected),
        },
        "companies": {
            "inbox": inbox,
            "review": reviewed,
            "qualified": qualified,
            "rejected": rejected,
        },
    }
    return resp

    # search_key = "AXZ-144RV"
    # company_keys = ds.keys(f"searches/{search_key}/companies")
    # all_domains = set([k.split("/")[-1] for k in company_keys])

    # event_keys = ds.keys(f"searches/{search_key}/events")
    # events = pd.DataFrame([ds[k] for k in event_keys])
    # # print(events)

    # evaluated = set()#set(events["domain"])

    # remaining_domains = list(all_domains - evaluated)[0:10]

    # companies = [ds[k] for k in company_keys if k.split("/")[-1] in remaining_domains]
    # # df = pd.DataFrame(companies)
    # # df['employee_count'] = df['employees'].apply(lambda x: x.get("value"))
    # # df = df.sort_values("employee_count", ascending=False)

    # resp = {
    #     "search_key": search_key,
    #     "actor_key": actor_key,
    #     "meta": {
    #         "inbox": len(companies),
    #         "reviewed": 17,
    #         "qualified": 3,
    #         # "...": len(events),
    #     },
    #     "companies": {
    #         "inbox": companies,
    #         "reviewed": [],
    #         "qualified": [],
    #     },
    # }
    # return resp



import os
import random
from dataclasses import asdict, dataclass, field
from hashlib import md5
from time import time

import pandas as pd
from dacite import from_dict
from twilio.rest import Client

from gandai.datastore import Cloudstore
from gandai.models.User import user_exists

ds = Cloudstore()
twilio_client = Client(
    ds["env/TWILIO_APP"], 
    ds["env/TWILIO_TOKEN"]
)


@dataclass
class Auth:
    key: str = field(init=False)
    phone: str
    code: int  # 6 digit
    expires: int = field(init=False)
    token: str = field(init=False)

    def __post_init__(self):
        assert len(self.code) == 6
        self.expires = int(time()) + (7 * 86400)
        self.key = f"auth/{self.phone}/{self.expires}"
        self.token = md5(self.key.encode()).hexdigest()


def _send_code(auth: Auth) -> None:
    message = twilio_client.messages.create(
        to=auth.phone,
        from_=ds["env/TWILIO_NUMBER"],
        body=f"Your Login Code is: {auth.code}",
    )
    print(f"Login Sent to {auth.phone}")


def send_code(phone: str) -> int:
    # should 
    # a) anyone be able to do a 2FA?, OR
    # b) only exisiting users?
    if user_exists(phone):
        auth_code = random.randint(100000, 999999)
        auth = Auth(phone=phone, code=auth_code)
        _send_code(auth)
        ds[auth.key] = asdict(auth)
        return 200
    else:
        print("Nah")
        return 401


def authenticate(code: int) -> str:
    for k in ds.keys("auth"):
        auth = from_dict(Auth, ds[k])
        if code == auth.code and auth.expires > int(time()):
            return auth.token


def validate(token: str) -> bool:
    pass



import asyncio
import re
from collections import Counter

# import mapply
import pandas as pd
import plotly.express as px
from nltk import RegexpTokenizer, ngrams
from nltk.corpus import stopwords
from scipy.special import softmax
from scipy.stats import zscore
from sklearn.preprocessing import MinMaxScaler

from gandai.datastore import Cloudstore

ds = Cloudstore()

# mapply.init(n_workers=-1, chunk_size=1, max_chunks_per_worker=8, progressbar=False)


def dealcloud_query():
    df = pd.read_feather("data/dealcloud_export.feather")
    df.columns = [col.lower().replace(" ", "_") for col in df.columns]
    return df


def _to_tokens(txt: str) -> list:
    tokenizer = RegexpTokenizer(r"\w+")
    tokens = tokenizer.tokenize(txt)
    lower_tokens = [t.lower() for t in tokens]
    clean = [t for t in lower_tokens if t not in stopwords.words("english")]
    return clean


def _to_ngram(tokens: list, n=2):
    grams = list(ngrams(tokens, 2))
    return [" ".join(list(gram)) for gram in grams]


def _to_ngrams(txt: str) -> list:
    if txt:
        tokens = _to_tokens(txt)
        grams: list = tokens
        grams.extend(_to_ngram(tokens, 2))
        grams.extend(_to_ngram(tokens, 3))
        return grams
    else:
        return []


def _get_match_ngrams(row) -> list:
    grams = []
    grams.extend(_to_ngrams(row["company_name"]))
    grams.extend(_to_ngrams(row["domain"]))
    grams.extend(_to_ngrams(row["business_description"]))
    grams.extend(_to_ngrams(row["main_industry"]))
    return grams


def _clean_days_since(txt):
    if txt:
        return int(re.findall(r"\d+", txt)[0])


def _get_domain(url) -> str:
    try:
        tokens = []
        for el in url.split("."):
            tokens.extend(el.split("/"))
        tld_index = tokens.index("com")
        domain_tokens = tokens[tld_index - 1 : tld_index + 1]
        domain = ".".join(domain_tokens)
        return domain
    except Exception as e:
        # print(e)
        pass


def _get_tld(url) -> str:
    tlds = [
        ".com",
        ".net",
        ".us",
        ".io",
        ".ca",
        ".uk",
        ".org",
        ".ai",
        ".biz",
        ".co",
        ".md",
        ".de",
        ".pharmacy",
        ".agency",
        ".solutions",
        ".global",
        ".eu",
    ]
    for tld in tlds:
        if tld in url:
            return tld
    return "?"


def dealcloud_feature_query():
    return_cols = [
        "company_name",
        "type",
        # "website",
        "domain",
        "business_description",
        "main_industry",
        # "linkedin",
        # "employee_linkedin_range",
        # "year_founded",
        # "employees",
        # "import_date",
        # "products",
        # "services",
        # "coverage_status",
        "days_since_contact",
        "zip5",
        "lat",
        "lon",
        "_ngrams",
        "_rev",
        "_employees",
    ]

    df = dealcloud_query()
    print(len(df))
    df["days_since_contact"] = df["days_since_contact"].apply(_clean_days_since)
    df["domain"] = df["website"].apply(_get_domain)
    df["tld"] = df["website"].dropna().apply(_get_tld)
    df["zip5"] = df["usa_postal_code"].dropna().apply(lambda x: str(x)[0:5])

    census = pd.read_feather("data/acs_zip5.feather")[["zip5", "lat", "lon"]]
    df = df.merge(census, how="left")

    # only recent
    # df = df[df["coverage_status"] != "Never Contacted"]

    df["_rev"] = df["source_revenue_(m)"]
    df = df.dropna(subset=["_rev"])
    df = df[df["_rev"] > 0]
    df = df[df["_rev"] < (10**3)]

    df["_employees"] = df["employees"]
    df = df.dropna(subset=["domain"])
    print(len(df))

    print("ngrams enabled... will take a min...")
    # df['_ngrams'] = df.mapply(_get_match_ngrams, axis=1)

    # df = df.sort_values("days_since_contact")
    return df[return_cols].reset_index(drop=True)



import pandas as pd

from gandai.datastore import Cloudstore

ds = Cloudstore()


def companies_query(search_key: str) -> pd.DataFrame:
    keys = ds.keys(f"searches/{search_key}/companies")
    df = pd.DataFrame(ds.load_async(keys))
    df["employee_count"] = df["employees"].apply(lambda x: x.get("value"))
    df = df.sort_values("employee_count", ascending=False).reset_index(drop=True)
    return df[["name", "domain", "description", "employee_count"]]


def events_query(search_key: str) -> pd.DataFrame:
    keys = ds.keys(f"searches/{search_key}/events")
    return pd.DataFrame(ds.load_async(keys))
